#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def _import_all():
    import importlib
    import os

    for fn in os.listdir(os.path.dirname(__file__)):
        if fn.endswith('_pb2.py'):
            try:
                importlib.import_module('.' + os.path.splitext(fn)[0], __name__)
            except ImportError:
                break


_import_all()
del _import_all
